import { next } from "./utils.ts";
import { range } from "./Range.ts";

export function it<T>(iter: Iterable<T>) {
	return new Iter(iter);
}

export function* map<I, O>(iter: Iterable<I>, mapping: (item: I) => O): Generator<O, void, void> {
	for (const item of iter) yield mapping(item);
}

export function* filter<I>(iter: Iterable<I>, predicate: (item: I) => boolean): Generator<I, void, void> {
	for (const item of iter) if (predicate(item)) yield item;
}

export function filter_map<I, O>(
	iter: Iterable<I>,
	mapping: (item: I) => O,
): Generator<Exclude<O, null | undefined>, void, void> {
	const filtered = filter(map(iter, mapping), (item) => item !== null && item !== undefined);
	return filtered as Generator<Exclude<O, null | undefined>, void, void>;
}

export function all<I>(iter: Iterable<I>, predicate: (item: I) => boolean) {
	for (const item of iter) if (!predicate(item)) return false;
	return true;
}

export function* enumerate<T>(iterator: Iterable<T>): Generator<[number, T], void, void> {
	let index = 0;
	for (const item of iterator) yield [index++, item];
}

export function* zip<A, B>(a: Iterable<A>, b: Iterable<B>): Generator<[A, B], void, void> {
	const iter_a = a[Symbol.iterator](), iter_b = b[Symbol.iterator]();
	while (true) {
		const next_a = next(iter_a), next_b = next(iter_b);
		if (next_a === null || next_b === null) return;
		yield [next_a, next_b];
	}
}

export function* chunk<T>(iter: Iterable<T>, size: number, incomplete = true): Generator<T[], void, void> {
	let chunk = [] as T[];
	for (const item of iter) {
		if (chunk.length === size) {
			yield chunk;
			chunk = [];
		}
		chunk.push(item);
	}
	if (incomplete) yield chunk;
}

export function* window<T>(iter: Iterable<T>, size: number, incomplete = false): Generator<T[], void, void> {
	const window = [] as T[];
	for (const item of iter) {
		window.push(item);
		if (window.length > size) window.splice(0, 1);
		if (incomplete || window.length === size) yield [...window];
	}
	if (incomplete) {
		for (const _ in range(0, window.length)) {
			window.splice(0, 1);
			yield [...window];
		}
	}
}

export function* fill<I, O>(iter: Iterable<I>, value: O) {
	for (const _item of iter) yield value;
}

export function fold<I, O>(iter: Iterable<I>, init: O, reduction: (item: I, acc: O) => O) {
	let accumulator = init;
	for (const item of iter) accumulator = reduction(item, accumulator);
	return accumulator;
}

export class Iter<T> {
	iter;
	[Symbol.iterator] = () => this.iter[Symbol.iterator]();

	constructor(iter: Iterable<T>) {
		this.iter = iter;
	}

	map<O>(mapping: (item: T) => O) {
		return new Iter(map(this, mapping));
	}

	filter(predicate: (item: T) => boolean) {
		return new Iter(filter(this, predicate));
	}

	filter_map<O>(mapping: (item: T) => O) {
		return new Iter(filter_map(this, mapping));
	}

	enumerate() {
		return new Iter(enumerate(this));
	}

	zip<U>(other: Iterable<U>) {
		return new Iter(zip(this, other));
	}

	chunk(size: number, incomplete = true) {
		return new Iter(chunk(this, size, incomplete));
	}

	window(size: number, incomplete = false) {
		return new Iter(window(this, size, incomplete));
	}

	array() {
		return [...this];
	}

	array_it() {
		// deno-lint-ignore no-explicit-any
		const result: any = [...this];
		result.it = () => it(result);
		return result as T[] & { it: () => Iter<T> };
	}

	collect<O>(class_: new (it: Iterable<T>) => O): O {
		return new class_(this.iter);
	}

	all(predicate: (item: T) => boolean) {
		return all(this, predicate);
	}

	fill<O>(value: O) {
		return fill(this, value);
	}

	fold<O>(init: O, reduction: (item: T, acc: O) => O) {
		return fold(this, init, reduction);
	}
}

Deno.test("test_comon", async () => {
	const { assertEquals } = await import("https://deno.land/std@0.224.0/assert/assert_equals.ts");
	const items = it(range(0, 10))
		.map((e) => e * 3)
		.enumerate()
		.filter(([_, e]) => (e % 2) == 1)
		.zip(range(0, 10))
		.map(JSON.stringify)
		.collect(Set<string>);
	assertEquals(items, new Set(["[[1,3],0]", "[[3,9],1]", "[[5,15],2]", "[[7,21],3]", "[[9,27],4]"]));
});

Deno.test("test_chunk", async () => {
	const { assertEquals } = await import("https://deno.land/std@0.224.0/assert/assert_equals.ts");
	const items = it(range(0, 5)).chunk(2).array();
	assertEquals(items, [[0, 1], [2, 3], [4]]);
});

Deno.test("test_window", async () => {
	const { assertEquals } = await import("https://deno.land/std@0.224.0/assert/assert_equals.ts");
	const items = it(range(0, 5)).window(3).array();
	assertEquals(items, [[0, 1, 2], [1, 2, 3], [2, 3, 4]]);
	const items_incomplete = it(range(0, 3)).window(3, true).array();
	assertEquals(items_incomplete, [[0], [0, 1], [0, 1, 2], [1, 2], [2]]);
});
